/*
  Copyright (C) 2002 by Krzysztof Langner

  This code is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package roboplayer.body;

import java.net.InetAddress;
import java.net.DatagramSocket;
import java.net.DatagramPacket;
import java.net.SocketException;
import java.io.IOException;
import java.util.StringTokenizer;
import java.util.Vector;
import roboplayer.body.TickListener;


/**
 * This class represents robot body. It is used to communicate with the server.
 * This class is the only thing which should be changed
 * if we want to move the code to real robot.
 * @author Krzysztof Langner
 * @version 1.0
 */
public class Body implements java.lang.Runnable, RobotCommands
{
	//---------------------------------------------------------------------------
  /**
   * Standard constructor
   *  @param host Server host
   *  @param port server port
   *  @throws SocketException when we can't connect to the server
   */
  public Body(InetAddress host, int port) throws SocketException
  {
    this.socket = new DatagramSocket();
		this.host = host;
		this.port = port;
  }


	//---------------------------------------------------------------------------
  /**
   * Connect to server
   * @param teamName name of the team
   * @param goalie true if goalkeeper
   * @return true we received init message
   */
  public boolean init(String teamName, boolean goalie)
  {
    String command;
    command = "(init " + teamName + " (version 5.04)";

    this.teamName = teamName;

    // Check if this is goalie
    if(goalie)
      command += " (goalie)";
    else
      command += ")";

    send(command);

    // Wait for initialization information from server
    // If everything Ok start listening for message
    if(!waitForInitMessage())
      return false;

    stopThread = false;

    return true;
  }

	//---------------------------------------------------------------------------
  /**
   * Reconnect to the server
   * @param teamName name of the team
   * @param uniformNumber player number
   */
  public void reconnect(String teamName, int uniformNumber)
  {
    send("(reconnect " + teamName + " " + uniformNumber + ")");

    // Wait for initialization information from server
    // If everything Ok start listening for message
    if(waitForInitMessage())
    {
      stopThread = false;
    }
  }

	//---------------------------------------------------------------------------
  /**
   * Disconnect from the server
   */
  public void bye()
  {
    stopThread = true;
    send("(bye)");
  }


	//============================================================================
  // Properities
	//---------------------------------------------------------------------------
	/**
   * Player side
   * @return side
   */
  public char getSide()
  {
    return side;
  }

	//---------------------------------------------------------------------------
	/**
   * Player unform number
   * @return player number
   */
  public int getUniformNumber()
  {
    return uniformNumber;
  }


	//---------------------------------------------------------------------------
	/**
   * Player current speed
   * @return current player speed
   */
  public double getSpeed()
  {
    return speed;
  }

	//---------------------------------------------------------------------------
	/**
   * Player current stamina
   * @return player stamina
   */
  public double getStamina()
  {
    return stamina;
  }

	//---------------------------------------------------------------------------
	/**
   * Player current efford
   * @return player efford
   */
  public double getEfford()
  {
    return effort;
  }

	//---------------------------------------------------------------------------
	/**
   * Player head angle
   * @return head angle
   */
  public double getHeadAngle()
  {
    return headAngle;
  }

	//---------------------------------------------------------------------------
	/**
  * Get information about last turn command which was already send,
  * but which was not included in any visual info yet.
  * @return last turn angle
  */
  public double getLastTurnAngle()
  {
    return lastTurnAngle;
  }

	//============================================================================
  // RobotCommands interface
	//---------------------------------------------------------------------------
	/**
   * This function sends catch command to the server
   * @param dir catch direction
   */
	public void catchBall(double dir)
	{
		send("(catch " + Double.toString(dir) + ")");
	}

	//---------------------------------------------------------------------------
	/**
   * This function sends chage_view command to the server
   * @param angle view angle
   * @param quality view quality
   */
	public void changeView(String angle, String quality)
	{
		send("(change_view " + angle + " " + quality + ")");
	}

	//---------------------------------------------------------------------------
	/**
   * This function sends dash command to the server
   * @param power dash power
   */
	public void dash(double power)
	{
    if(power > 130)
      power = 130;
    else if(power < -30)
      power = -30;

//    System.out.println("(dash " + Double.toString(power) + ") time = " + lastTickTime);
		send("(dash " + Double.toString(power) + ")");
	}

	//---------------------------------------------------------------------------
	/**
   * This function sends kick command to the server
   * @param power kick power
   * @param direction kick direction
   */
	public void kick(double power, double direction)
	{
		send("(kick " + Double.toString(power) + " " + Double.toString(direction) + ")");
	}

	//---------------------------------------------------------------------------
	/**
   * This function sends move command to the server.
   * This function can only be used to position player before game.
   * @param x move robot to this direction
   * @param y move robot to this direction
   */
	public void move(double x, double y)
	{
		send("(move " + Double.toString(x) + " " + Double.toString(y) + ")");
	}

	//---------------------------------------------------------------------------
	/**
   * This function sends say command to the server
   * @param message message
   */
	public void say(String message)
	{
		send("(say " + message + ")");
	}

	//---------------------------------------------------------------------------
	/**
   * This function sends turn command to the server
   * @param angle turn angle
   */
	public void turn(double angle)
	{
		send("(turn " + Double.toString(angle) + ")");
    lastTurnAngle = angle;
//    System.out.println("(turn " + Double.toString(angle) + ")");
	}

	//---------------------------------------------------------------------------
	/**
   * This function sends turn head command to the server
   * @param angle turn angle
   */
	public void turnNeck(double angle)
	{
		send("(turn_neck " + Double.toString(angle) + ")");

		//System.out.println("(turn_neck " + Double.toString(angle) + ")");
    headAngle += angle;

    if(headAngle > 90)
      headAngle = 90;
    else if(headAngle < -90)
      headAngle = -90;
	}

	//============================================================================
  // Additional functions
	//---------------------------------------------------------------------------
	/**
   * This function unregisters TickListener interface
   * @param l listener
   */
  public void removeTickListener(TickListener l)
  {
    tickListeners.removeElement(l);
  }

	//---------------------------------------------------------------------------
	/**
   * This function registers TickListener interface
   * @param l listener
   */
  public void addTickListener(TickListener l)
  {
    tickListeners.addElement(l);
  }

	//---------------------------------------------------------------------------
	/**
   * This function unregisters SeeListener interface
   * @param l listener
   */
  public void removeSeeListener(SeeListener l)
  {
    seeListeners.removeElement(l);
  }

	//---------------------------------------------------------------------------
	/**
   * This function registers SeeListener interface
   * @param l listener
   */
  public synchronized void addSeeListener(SeeListener l)
  {
    seeListeners.addElement(l);
  }

	//---------------------------------------------------------------------------
	/**
   * This function unregisters HearListener interface
   * @param l listener
   */
  public synchronized void removeHearListener(HearListener l)
  {
    hearListeners.removeElement(l);
  }

	//---------------------------------------------------------------------------
	/**
   * This function registers HearListener interface
   * @param l listener
   */
  public synchronized void addHearListener(HearListener l)
  {
    hearListeners.contains(l);
  }

  //////////////////////////////////////////////////////////////////////////////
  // Implementation
  //////////////////////////////////////////////////////////////////////////////
	//---------------------------------------------------------------------------
  /**
   * Send message to the server
   * @param message message to send
   */
  private void send(String message)
  {
    byte[]  data = message.getBytes();
		DatagramPacket packet = new DatagramPacket(data, data.length, host, port);

    //System.out.println(message);
		try
    {
			socket.send(packet);
		}
    catch(IOException e)
    {
			System.err.println("socket sending error " + e);
      System.exit(-1);
		}

    lastTurnAngle = 0;
  }

	//---------------------------------------------------------------------------
	/**
   * This function waits for new message from server
   * @return      Received string
   */
	private String receive()
	{
		byte[] buffer = new byte[MSG_SIZE];
		DatagramPacket packet = new DatagramPacket(buffer, MSG_SIZE);

		try{
			socket.receive(packet);
		}catch(IOException e){
			System.err.println("socket receiving error " + e);
		}

		return new String(buffer);
	}

	//---------------------------------------------------------------------------
	/**
   * This function waits for initializations message from the server.
   * @return      The resultant of initialization
   */
  private boolean waitForInitMessage()
  {
		byte[] buffer = new byte[MSG_SIZE];
		DatagramPacket  packet = new DatagramPacket(buffer, buffer.length);
    String          msg;
    StringTokenizer	tokenizer;
    String          token;

		try
    {
      while(true)
      {
			  socket.receive(packet);
        msg = new String(packet.getData());
        tokenizer = new StringTokenizer(msg,"() ");
        token = tokenizer.nextToken();

        if( token.compareTo("init") == 0)
        {
          port = packet.getPort();
          side = tokenizer.nextToken().charAt(0);
          uniformNumber = Integer.parseInt(tokenizer.nextToken());
          // Send as a hear message
          msg = "(hear 0 referee " + tokenizer.nextToken() + ")";
          sendHearMessage(msg);

          break;
        }
        else if( token.compareTo("error") == 0)
        {
          System.err.println(msg);
          return false;
        }
      }

		}catch(IOException e){
			System.err.println("socket receiving error " + e);
		}

		return true;
  }

	//---------------------------------------------------------------------------
	/**
  * This function process sense_body message.
  * It is also used to synchronize visual information.
  * @param msg message
  */
  private void senseBody(String msg)
  {
    StringTokenizer	tokenizer = new StringTokenizer(msg,"() ", true);
    String          token;
    long            time;

    // Skip first "("
    tokenizer.nextToken();
    // Get "sense_body"
    token = tokenizer.nextToken();
    if(token.compareTo("sense_body") != 0)
      return;

    // Skip space
    tokenizer.nextToken();
    // Skip TIME
    time = Integer.parseInt(tokenizer.nextToken());
    sendTickMessage(time);

    while(tokenizer.hasMoreTokens())
    {
      // Check for closing bracket
      token = tokenizer.nextToken();
      if(token.compareTo(")") == 0)
        break;

      // Get "("
      tokenizer.nextToken();
      // Get info type
      token = tokenizer.nextToken();
      // Remove first space
      tokenizer.nextToken(); // space
      if(token.compareTo("view_mode") == 0)
      {
        // Skip this information
        token = tokenizer.nextToken(); // high|low
        token = tokenizer.nextToken(); // space
        token = tokenizer.nextToken(); // narrow|normal|wide
      }
      else if(token.compareTo("stamina") == 0)
      {
        stamina = Float.valueOf(tokenizer.nextToken()).floatValue();
        tokenizer.nextToken(); // space
        effort = Float.valueOf(tokenizer.nextToken()).floatValue();
      }
      else if(token.compareTo("speed") == 0)
      {
        speed = Float.valueOf(tokenizer.nextToken()).floatValue();
      }
      else if(token.compareTo("kick") == 0)
      {
        tokenizer.nextToken(); // kick_count
      }
      else if(token.compareTo("dash") == 0)
      {
        tokenizer.nextToken(); // dash_count
      }
      else if(token.compareTo("turn") == 0)
      {
        tokenizer.nextToken(); // turn_count
      }
      else if(token.compareTo("say") == 0)
      {
        tokenizer.nextToken(); // say_count
      }
      else if(token.compareTo("turn_neck") == 0)
      {
        tokenizer.nextToken(); // turn_neck_count
      }
      else if(token.compareTo("head_angle") == 0)
      {
        headAngle = Float.valueOf(tokenizer.nextToken()).floatValue();
//        System.out.println("headAngle = " + headAngle);
      }

      // Get closing bracket ")"
      token = tokenizer.nextToken();
    }

//    if(stamina < 100)
//      System.out.println("Speed = " + speed + "Stamina = " + stamina);
  }

	//---------------------------------------------------------------------------
  /**
   * Main loop of the robot.
   * Wait for information from SoccerServer and send it to registered listeners.
   *  How each information is processed:
   *    - sense_body is processed internally
   *    - see is send to each registered SeeListener interface
   *    - hear is send to each registered HearListener interface
   *
   */
  public void run()
  {
    String          msg;
    StringTokenizer	tokenizer;
    String          token;

    while(!stopThread)
    {
      msg = receive().trim();
      tokenizer = new StringTokenizer(msg,"() ");
      token = tokenizer.nextToken();

      if(token.compareTo("see") == 0){
        lastTurnAngle = 0;
        sendSeeMessage(msg);
        sendTickMessage(lastTickTime);
      }
      else if(token.compareTo("hear") == 0){
        sendHearMessage(msg);
      }
      else if(token.compareTo("sense_body") == 0){
        senseBody(msg.trim());
      }
      else
        System.err.println(msg);
    }
  }

	//---------------------------------------------------------------------------
  /**
   * This function sends see message to registered interfaces
   * @param msg message
   */
  private void sendSeeMessage(String msg)
  {
    int count = seeListeners.size();
    for(int i = 0; i < count; i++)
      ((SeeListener) seeListeners.elementAt(i)).onSee(msg);
  }

	//---------------------------------------------------------------------------
  /**
   * This function sends see message to registered interfaces
   * @param time time which passed form last function call
   */
  private void sendTickMessage(long time)
  {
    if(lastTickTime < time)
    {
//      System.out.println("stamina = " + stamina);
      lastTickTime = time;
      int count = tickListeners.size();
      for(int i = 0; i < count; i++)
        ((TickListener) tickListeners.elementAt(i)).onTick(time);
    }
  }

	//---------------------------------------------------------------------------
  /**
   * This function sends hear message to registered interfaces
   * @param msg message
   */
  private void sendHearMessage(String msg)
  {
    int count = hearListeners.size();
    for(int i = 0; i < count; i++)
      ((HearListener) hearListeners.elementAt(i)).OnHear(msg);
  }

  //////////////////////////////////////////////////////////////////////////////
  // Private members
  //////////////////////////////////////////////////////////////////////////////

  /** Team name */
  private String  teamName;
  /** uniform number */
  private int     uniformNumber;
  /** player side */
  private char    side;
  /** player current speed */
  private double speed;
  /** player stamina */
  private double stamina;
  /** player effort */
  private double effort;
  /** neck angle */
  private double headAngle;
  /** Last turn angle */
  private double lastTurnAngle = 0;

  /** Socket to communicate with the server */
  private DatagramSocket  socket;
  /** Server address*/
  private InetAddress	    host;
  /** Server port*/
  private int		          port;
  /** Tick listeners */
  private Vector tickListeners = new Vector();
  /** See listeners */
  private Vector seeListeners = new Vector();
  /** Hear listeners */
  private Vector hearListeners = new Vector();
  /** Stops thread */
 	private boolean	stopThread;
  /** Last tick time */
  private long    lastTickTime = -1;



	// constants
  /** Size of socket buffer*/
	private static final int	MSG_SIZE = 2048;

}

