/*
  Copyright (C) 2002 by Krzysztof Langner

  This code is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package roboplayer.expert;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.FileReader;
import java.util.Vector;
import java.util.Enumeration;
import java.util.StringTokenizer;
import roboplayer.body.TickListener;
import roboplayer.action.Actions;
import roboplayer.query.QuerySystem;

/**
 * Base class for various experts.
 * Objects of this class make decision about what to do.
 * @author Krzysztof Langner
 * @version 1.0
 */
public class Expert implements TickListener
{
	//---------------------------------------------------------------------------
	/** This constructor */
	public Expert(Actions actions,  QuerySystem queries)
	{
    this.queries = queries;
    this.actions = actions;
	}

	//---------------------------------------------------------------------------
	/**
	* Load logic from file
	*/
	public void loadFile(String fileName)
	{
    BufferedReader  in;
    String          line;
    StringTokenizer tokenizer;


    // read info from file
    try
    {
      in = new BufferedReader(new FileReader(fileName));
      // In the first line expect team name
      while((line = in.readLine()) != null){
        line.trim();
        // check if this is comment
        if(line.length() > 0 && line.charAt(0) == '#')
          continue;

        tokenizer = new StringTokenizer(line, ":", false);
        // check if this is valid command
        if(tokenizer.countTokens() == 2){
          rules.add(new Rule(
            tokenizer.nextToken().trim(),
            tokenizer.nextToken().trim(),
            queries));
        }
      }
 		}
    catch(IOException e){
			System.err.println("Error processing file\n");
			System.err.println(e.getMessage());
      System.exit(-1);
    }

    // Prepare to game
    prepareToGame();

  }


	//---------------------------------------------------------------------------
	/** This is main Brain function used to make decision
	* Check play mode and call suitable function
	*/
	public void onTick(long time)
	{

    try
    {
      if(!queries.ask("isTimeOver"))
      {
        if(queries.ask("isPlayOn"))
        {
          if(queries.ask("isPositionEstimated"))
          	play();
        }
        else
        {
          beforePlayOn();
        }
      }
    }
    catch(Exception e)
    {
      System.out.println(e.toString());
      e.printStackTrace();
    }
  }

//===========================================================================
// Here are supporting functions

	//---------------------------------------------------------------------------
	/** Play game */
  protected void prepareToGame()
  {
    actions.execute("PrepareToGame");
  }

	//---------------------------------------------------------------------------
	/**
	 *  Choose what to do.
  */
  protected void play()
  {
    Enumeration e = rules.elements();
    String action;

    while(e.hasMoreElements()){
      Rule rule = (Rule)e.nextElement();
      action = rule.execute();
      if(action != null){
//System.out.println(action);
        actions.execute(action);
        break;
      }
    }
  }


	//---------------------------------------------------------------------------
	/** Small break */
  protected void beforePlayOn()
  {
    play();
  }

  //===========================================================================
  // Protected members
  /** place where all visual information is stored */
	protected QuerySystem   queries;
  /** Player's skills */
  protected Actions       actions;
  /** Rules */
  private Vector          rules = new Vector();

}

