/*
  Copyright (C) 2002 by Krzysztof Langner

  This code is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package roboplayer.memory;

import java.util.Enumeration;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.Hashtable;
import roboplayer.body.Settings;
import roboplayer.geometry.Vector2d;
import roboplayer.geometry.Polar;
import roboplayer.geometry.Angle;
import roboplayer.body.Body;
import roboplayer.body.SeeListener;
import roboplayer.body.TickListener;

/**
 * This class represents visual memory.
 * It not only read information from sensors, but it can be used to
 * store information and estimate position for objects out of the view
 * @author Krzysztof Langner
 * @version 1.0
 */
public class VisualInfo implements SeeListener, TickListener
{
	//---------------------------------------------------------------------------
	// This constructor:
	// - initializes all variables
	public VisualInfo(Body body, char side, String team)
	{
    m_body = body;
		m_side = side;
		m_team = team;

    initLandmarks();
    m_body.addSeeListener(this);
    m_body.addTickListener(this);
  }

	//---------------------------------------------------------------------------
	// This function parses sensor information
	public synchronized void onSee(String info)
	{
	  VisualInfoParser	parser = new VisualInfoParser(info, m_body.getHeadAngle());

  	parser.parse();
	  store(parser.getData());
  	estimatePosition(parser.getData());
    infoTime = parser.getTime();

//    System.out.println("time = " + infoTime + " -------------------------------");
//    System.out.println(m_direction);
//    if(ball != null && ball.probability == 1)
//      System.out.println("ball dir = " + ball.dir);
	}


	//---------------------------------------------------------------------------
  /**
   * Update visual information. Time has changed.
   */
  public synchronized void onTick(long time) {

    // If visual information is old then aproximate it.
    if(infoTime < time-1){
      updateInfo();
    }
  }

	//---------------------------------------------------------------------------
	// This function returns my direction
	public synchronized double getDirection()
	{
		return m_direction;
	}


	//---------------------------------------------------------------------------
	// This function returns my direction
	public synchronized Vector2d getPosition()
	{
		return m_pos;
	}


	//---------------------------------------------------------------------------
	// This function looks for a ball
	public synchronized ObjectInfo getBall()
	{
		return ball;
	}


	//---------------------------------------------------------------------------
	// This function looks for opponent goal
	public synchronized Polar getOpponentGoalPos()
	{
		return getLandmark("goal r");
	}


	//---------------------------------------------------------------------------
	// This function looks for opponent goal
	public synchronized Polar getMyGoalPos()
	{
		return getLandmark("goal l");
	}


	//---------------------------------------------------------------------------
	/** This function calculates position from polar coordinates */
	public synchronized Vector2d positionFromPolar(Polar p)
	{
    double    dir = Angle.Deg2Rad(p.dir + m_direction);
    double    dist = p.dist;
    Vector2d  pos = new Vector2d();

    pos.x = Math.cos(dir)*dist + m_pos.x;
    pos.y = Math.sin(dir)*dist + m_pos.y;

    return pos;
	}


	//---------------------------------------------------------------------------
	/** This function calculates position from polar coordinates */
	public synchronized Polar nextPolar(ObjectInfo obj)
	{
    Vector2d p1, p2;

    p1 = positionFromPolar(obj);
    p2 = obj.getChange().getPosition();

//    System.out.println("Change = " + p2);
    p1.add(p2);

    return getPolar(p1);
	}


	//---------------------------------------------------------------------------
	// This function looks for teammate
	public synchronized ObjectInfo getNearestTeammate()
	{
		ObjectInfo teammate = null;

		for(int c = 0 ; c < m_players.size() ; c ++)
		{
			ObjectInfo player = (ObjectInfo)m_players.elementAt(c);
			StringTokenizer tokenizer = new StringTokenizer(player.name, " ");
			// First token is name
			tokenizer.nextToken();
			// Then we can find team name
			if(tokenizer.hasMoreTokens() &&
					tokenizer.nextToken().compareTo(m_team) == 0)
			{
				if(teammate == null || player.dist < teammate.dist)
					teammate = player;
			}
		}

		return teammate;
	}


	//---------------------------------------------------------------------------
	// This function looks for nearest player to specified position
	public synchronized ObjectInfo getNearestPlayer(Polar pos)
	{
		ObjectInfo nearest = null;
		double dist;

		if( ball == null )
			return null;

		dist = ball.dist;

		for(int c = 0 ; c < m_players.size() ; c ++)
		{
			ObjectInfo player = (ObjectInfo)m_players.elementAt(c);
			double d = player.getDistance(ball);
			if( d < dist )
			{
				nearest = player;
				dist = d;
			}

		}

		return nearest;
	}


	//---------------------------------------------------------------------------
	/**
   * This function looks for nearest player to specified position.
   * @param pos Distance is calculated from this position
   * @return Player object or null if we are the nearest player
   */
	public synchronized ObjectInfo getNearestTeammate(Polar pos)
	{
		ObjectInfo nearest = null;
		double dist = -1;

		for(int c = 0 ; c < m_players.size() ; c ++)
		{
			ObjectInfo player = (ObjectInfo)m_players.elementAt(c);
			StringTokenizer tokenizer = new StringTokenizer(player.name, " ");
			// First token is name
			tokenizer.nextToken();
			// Then we can find team name
			if(tokenizer.hasMoreTokens() &&
					tokenizer.nextToken().compareTo(m_team) == 0)
			{
				double d = player.getDistance(pos);
				if( nearest == null || d < dist )
				{
					nearest = player;
					dist = d;
				}
			}
		}

		return nearest;
	}


	//---------------------------------------------------------------------------
	// This function looks for opponent player at specified direction
	public synchronized ObjectInfo getOpponentPlayer(double dir)
	{
		for(int c = 0 ; c < m_players.size() ; c ++)
		{
			ObjectInfo player = (ObjectInfo)m_players.elementAt(c);
			if(Math.abs(player.dir - dir) < 5 &&
				! player.name.regionMatches(0,m_team,0,m_team.length()))
			{
				return player;
			}
		}

		return null;
	}


	//---------------------------------------------------------------------------
	// This function looks for landmark
	public synchronized Polar getLandmark(String name)
	{
		Vector2d pos = (Vector2d)m_landmarks.get(name);

		if(pos != null)
			return getPolar(pos);

		return null;
	}


	//---------------------------------------------------------------------------
	// This function looks for landmark
	public double getTurnAngle()
	{
		return m_turnAngle;
	}


	//---------------------------------------------------------------------------
	/** This function returns polar coordinates of specified point */
	public Polar getPolar(Vector2d pos)
	{
		Polar polar = new Polar();

		polar.dist = Math.sqrt( Math.pow((pos.x-m_pos.x),2) + Math.pow((pos.y-m_pos.y),2) );

		double angle = Math.atan((pos.y-m_pos.y)/(pos.x-m_pos.x));
		polar.dir = Angle.Rad2Deg(angle);
		if( pos.x < m_pos.x )
		{
			if(polar.dir > 0)
				polar.dir = -180 + polar.dir;
			else
				polar.dir = 180 + polar.dir;
		}

		polar.dir = polar.dir - m_direction;

		if(polar.dir < -180)
			polar.dir = 360 + polar.dir;
		else if(polar.dir > 180)
			polar.dir = -360 + polar.dir;

		return polar;
	}


//===========================================================================
// Private implementation

	//---------------------------------------------------------------------------
	/** This function puts see information into our memory */
	private void store(Enumeration enum)
	{
		m_players.removeAllElements();
    if(ball != null)
      ball.probability = 0.5 * ball.probability;

    while(enum.hasMoreElements())
		{
			ObjectInfo object = (ObjectInfo)enum.nextElement();
			if(object.name.compareTo("ball") == 0)
				ball = object;
			else if(object.name.charAt(0) == 'p')	// player
				m_players.addElement(object);
		}
	}

	//---------------------------------------------------------------------------
	/**
  * This function updates state from old state.
  * Currently only ball is estimated and we don't care about our speed.
  */
	private void updateInfo()
	{
    if(ball != null){
      ball.probability = 0.5 * ball.probability;
      // Update turn & speed
//System.out.println("-------------------------------------");
//System.out.println("speed = " + m_body.getSpeed());
//System.out.println(positionFromPolar(ball));
      ball.dir = ball.dir - m_body.getLastTurnAngle();
      m_direction += m_body.getLastTurnAngle();
//System.out.println(m_body.getLastTurnAngle());
//System.out.println(positionFromPolar(ball));
    }
	}

	//---------------------------------------------------------------------------
  /** Initialize landmarks */
  private void initLandmarks()
  {
		// Initialize landmarks
		m_landmarks.put(new String("goal l"), new Vector2d(-Settings.PITCH_LENGTH/2, 0));
		m_landmarks.put(new String("goal r"), new Vector2d(Settings.PITCH_LENGTH/2, 0));

    // Init flags
		m_landmarks.put(new String("flag l t"), new Vector2d(	-Settings.PITCH_LENGTH/2,
																	              -Settings.PITCH_WIDTH/2));
		m_landmarks.put(new String("flag l b"),
										new Vector2d(	-Settings.PITCH_LENGTH/2,
																	Settings.PITCH_WIDTH/2));
		m_landmarks.put(new String("flag c"),
										new Vector2d(0, 0));
		m_landmarks.put(new String("flag c t"),
										new Vector2d(0, -Settings.PITCH_WIDTH/2));
		m_landmarks.put(new String("flag c b"),
										new Vector2d(0, Settings.PITCH_WIDTH/2));
		m_landmarks.put(new String("flag r t"),
										new Vector2d(	Settings.PITCH_LENGTH/2,
																	-Settings.PITCH_WIDTH/2));
		m_landmarks.put(new String("flag r b"),
										new Vector2d(Settings.PITCH_LENGTH/2,
																 Settings.PITCH_WIDTH/2));

		m_landmarks.put(new String("flag p l t"),
										new Vector2d(	-Settings.PITCH_LENGTH/2+Settings.PENALTY_AREA_LENGTH,
																	-Settings.PENALTY_AREA_WIDTH/2));
		m_landmarks.put(new String("flag p l c"),
										new Vector2d(-Settings.PITCH_LENGTH/2+Settings.PENALTY_AREA_LENGTH, 0));
		m_landmarks.put(new String("flag p l b"),
										new Vector2d(-Settings.PITCH_LENGTH/2+Settings.PENALTY_AREA_LENGTH,
																	Settings.PENALTY_AREA_WIDTH/2));
		m_landmarks.put(new String("flag p r t"),
										new Vector2d(	Settings.PITCH_LENGTH/2-Settings.PENALTY_AREA_LENGTH,
																	-Settings.PENALTY_AREA_WIDTH/2));
		m_landmarks.put(new String("flag p r c"),
										new Vector2d(Settings.PITCH_LENGTH/2-Settings.PENALTY_AREA_LENGTH, 0));
		m_landmarks.put(new String("flag p r b"),
										new Vector2d(	Settings.PITCH_LENGTH/2-Settings.PENALTY_AREA_LENGTH,
																	Settings.PENALTY_AREA_WIDTH/2));
		m_landmarks.put(new String("flag g l t"),
										new Vector2d( -Settings.PITCH_LENGTH/2, -Settings.getGoalWidth()/2));
		m_landmarks.put(new String("flag g l b"),
										new Vector2d(	-Settings.PITCH_LENGTH/2, Settings.getGoalWidth()/2));
		m_landmarks.put(new String("flag g r t"),
										new Vector2d( Settings.PITCH_LENGTH/2, -Settings.getGoalWidth()/2));
		m_landmarks.put(new String("flag g r b"),
										new Vector2d(	Settings.PITCH_LENGTH/2, Settings.getGoalWidth()/2));

    // Out of the field flags
		m_landmarks.put(new String("flag l t 30"), new Vector2d(	-Settings.PITCH_LENGTH/2 - 5, -30));
		m_landmarks.put(new String("flag l t 20"), new Vector2d(	-Settings.PITCH_LENGTH/2 - 5, -20));
		m_landmarks.put(new String("flag l t 10"), new Vector2d(	-Settings.PITCH_LENGTH/2 - 5, -10));
		m_landmarks.put(new String("flag l 0"), new Vector2d(	-Settings.PITCH_LENGTH/2 - 5, 0));
		m_landmarks.put(new String("flag l b 10"), new Vector2d(	-Settings.PITCH_LENGTH/2 - 5, 10));
		m_landmarks.put(new String("flag l b 20"), new Vector2d(	-Settings.PITCH_LENGTH/2 - 5, 20));
		m_landmarks.put(new String("flag l b 30"), new Vector2d(	-Settings.PITCH_LENGTH/2 - 5, 30));

		m_landmarks.put(new String("flag r t 30"), new Vector2d( Settings.PITCH_LENGTH/2 + 5, -30));
		m_landmarks.put(new String("flag r t 20"), new Vector2d( Settings.PITCH_LENGTH/2 + 5, -20));
		m_landmarks.put(new String("flag r t 10"), new Vector2d( Settings.PITCH_LENGTH/2 + 5, -10));
		m_landmarks.put(new String("flag r 0"), new Vector2d(	Settings.PITCH_LENGTH/2 + 5, 0));
		m_landmarks.put(new String("flag r b 10"), new Vector2d( Settings.PITCH_LENGTH/2 + 5, 10));
		m_landmarks.put(new String("flag r b 20"), new Vector2d( Settings.PITCH_LENGTH/2 + 5, 20));
		m_landmarks.put(new String("flag r b 30"), new Vector2d( Settings.PITCH_LENGTH/2 + 5, 30));

		m_landmarks.put(new String("flag t l 50"), new Vector2d( -50, - Settings.PITCH_WIDTH/2 - 5));
		m_landmarks.put(new String("flag t l 40"), new Vector2d( -40, - Settings.PITCH_WIDTH/2 - 5));
		m_landmarks.put(new String("flag t l 30"), new Vector2d( -30, - Settings.PITCH_WIDTH/2 - 5));
		m_landmarks.put(new String("flag t l 20"), new Vector2d( -20, - Settings.PITCH_WIDTH/2 - 5));
		m_landmarks.put(new String("flag t l 10"), new Vector2d( -10, - Settings.PITCH_WIDTH/2 - 5));
		m_landmarks.put(new String("flag t 0"), new Vector2d( 0, - Settings.PITCH_WIDTH/2 - 5));
		m_landmarks.put(new String("flag t r 10"), new Vector2d( 10, - Settings.PITCH_WIDTH/2 - 5));
		m_landmarks.put(new String("flag t r 20"), new Vector2d( 20, - Settings.PITCH_WIDTH/2 - 5));
		m_landmarks.put(new String("flag t r 30"), new Vector2d( 30, - Settings.PITCH_WIDTH/2 - 5));
		m_landmarks.put(new String("flag t r 40"), new Vector2d( 40, - Settings.PITCH_WIDTH/2 - 5));
		m_landmarks.put(new String("flag t r 50"), new Vector2d( 50, - Settings.PITCH_WIDTH/2 - 5));

		m_landmarks.put(new String("flag b l 50"), new Vector2d( -50, Settings.PITCH_WIDTH/2 + 5));
		m_landmarks.put(new String("flag b l 40"), new Vector2d( -40, Settings.PITCH_WIDTH/2 + 5));
		m_landmarks.put(new String("flag b l 30"), new Vector2d( -30, Settings.PITCH_WIDTH/2 + 5));
		m_landmarks.put(new String("flag b l 20"), new Vector2d( -20, Settings.PITCH_WIDTH/2 + 5));
		m_landmarks.put(new String("flag b l 10"), new Vector2d( -10, Settings.PITCH_WIDTH/2 + 5));
		m_landmarks.put(new String("flag b 0"),    new Vector2d( 0,   Settings.PITCH_WIDTH/2 + 5));
		m_landmarks.put(new String("flag b r 10"), new Vector2d( 10,  Settings.PITCH_WIDTH/2 + 5));
		m_landmarks.put(new String("flag b r 20"), new Vector2d( 20,  Settings.PITCH_WIDTH/2 + 5));
		m_landmarks.put(new String("flag b r 30"), new Vector2d( 30,  Settings.PITCH_WIDTH/2 + 5));
		m_landmarks.put(new String("flag b r 40"), new Vector2d( 40,  Settings.PITCH_WIDTH/2 + 5));
		m_landmarks.put(new String("flag b r 50"), new Vector2d( 50,  Settings.PITCH_WIDTH/2 + 5));

	}

	//---------------------------------------------------------------------------
  /** this function scans field */
  private void scanField()
  {
    double headAngle =  m_body.getHeadAngle();

    // Scan
//    System.out.println("Scan");
    if(m_scanAreaID == 5)
    {
      m_turnAngle = 180;
      m_scanAreaID = 0;
    }
    else
    {
      m_body.turnNeck(-headAngle-60 + m_scanAreaID*30);
      m_scanAreaID ++;
    }
  }

	//---------------------------------------------------------------------------
	// This function calculates current position and direction based
	// on given as a parameter object
	//
	private boolean estimatePosition(Enumeration enum)
	{
		ObjectInfo line = null;
		ObjectInfo obj1 = null;
		ObjectInfo obj2 = null;
    ObjectInfo oldBall = ball;
    Vector2d   ballPos = null;

    if(ball != null && ball.probability < 1)
      ballPos = positionFromPolar(ball);

		// first try to get landmarks which we will use
		// to estimate our position
		while(enum.hasMoreElements())
		{
			ObjectInfo object = (ObjectInfo)enum.nextElement();

			if (object.name.charAt(0) == 'f' || object.name.charAt(0) == 'g')
			{
				if(obj1 == null)
        {
          obj1 = object;
        }
        else if(obj1.dist > object.dist)
				{
					obj2 = obj1;
					obj1 = object;
				}
        else if(obj2 == null)
        {
          obj2 = object;
        }
        else if(obj2.dist > object.dist)
				{
					obj2 = object;
				}
			}
			else if(object.name.charAt(0) == 'l')
			{
				if(line == null || object.dist > line.dist)
            line = object;
			}
		}
    // !!! Blad w Soccer serverze
		if(line != null)
      line.dir = Angle.Normalize(line.dir - m_body.getHeadAngle());
    // !!! Blad w Soccer serverze

    if(obj1 == null)
    {
      System.out.println("******** I don't have any marker ********");

      // Don't update position
		  if(line != null)
        estimateDirection(line);
    }
    else if(obj2 == null)
    {
		  if(line != null)
      {
        estimateDirection(line);
        positionFromFlagAndLine(obj1, line);
      }
      else
      {
        System.out.println("******** I don't have marker and line********");
      }
    }
    else
    {
      //System.out.println("2 objects");
      positionFromFlags(obj1, obj2);
      // It is better to estimate direction from line
//		  if(line != null)
//        estimateDirection(line);
    }

    if(ballPos != null)
    {
      Polar p = getPolar(ballPos);
      ball.dist = p.dist;
      ball.dir = p.dir;
    }

//    System.out.println("Vector2d = " + m_pos);
//    System.out.println("Direction = " + m_direction );
		return true;
	}

	//---------------------------------------------------------------------------
	// This function calculates current direction
	//
	private void estimateDirection(ObjectInfo line)
	{
    //System.out.println(line.name + " dir = " + line.dir);

		switch (line.name.charAt(5))
		{
		case 't' :
			m_direction = -line.dir - ((line.dir < 0) ? 180.0 : 0.0) ;
			break ;

		case 'b' :
			m_direction = -line.dir - ((line.dir > 0) ? -180.0 : 0.0) ;
			break ;

		case 'l' :
			m_direction = -line.dir - ((line.dir < 0) ? -90.0 : 90.0) ;
			break ;

		case 'r' :
			m_direction = -line.dir - ((line.dir > 0) ? -90.0 : 90.0) ;
			break ;

		}

		if (m_side == 'r')
			m_direction += ((m_direction > 0) ? -180.0 : 180.0);

	}

	//---------------------------------------------------------------------------
	// This function calculates current position
	private void positionFromFlags(ObjectInfo obj1, ObjectInfo obj2)
	{
    double dist, dist2;
    double x1, x2, y1, y2, r1, r2, a, h, x3, y3;
    double ratio;

		Vector2d pos1 = (Vector2d)m_landmarks.get(obj1.name);
		Vector2d pos2 = (Vector2d)m_landmarks.get(obj2.name);
		if(pos1 == null || pos2 == null)
    {
      System.out.println("*** ??????????? ***  " + obj1.name);
      System.out.println("*** ??????????? ***  " + obj2.name);
			return;
    }

//System.out.println(obj1.name + " (" + pos1.x + ", " + pos1.y + ")");
//System.out.println(obj2.name + " (" + pos2.x + ", " + pos2.y + ")");
  	// do the calculations
	  r1 = obj1.dist;
  	r2 = obj2.dist;

	  x1 = pos1.x;
  	x2 = pos2.x;

	  y1 = pos1.y;
  	y2 = pos2.y;

	  // calculate the square distance of the two flags
  	dist2 = (x2 - x1)*(x2 - x1) + (y2 - y1)*(y2 - y1);
	  dist = Math.sqrt(dist2);
  	if (dist > r1 + r2)
	  {
		  // the circles would not intersect (because of inaccuracy)
  		dist = r1 + r2;
	  	dist2 = dist * dist;
  	}
	  else if ((r1 > r2) && (dist + r2 < r1))
  	{
	  	// the circles would not intersect (inaccuracy)
		  dist = r1 - r2;
  		dist2 = dist * dist;
	  }
  	else if ((r2 > r1) && (dist + r1 < r2))
	  {
		  // the circles would not intersect
  		dist = r2 - r1;
	  	dist2 = dist * dist;
  	}

	  r1 *= r1;
  	r2 *= r2;

  	// a = (r1^2 - r2^2 + d^2 ) / (2 d)
	  a = (r1 - r2 + dist2) / (2.0 * dist);
  	// h^2 = r1^2 - a^2
	  h = r1 - a*a;
  	if (h < 0.0)
	  	h = 0.0;
  	else
	  	h = Math.sqrt(h);

  	// P3 = P1 + a ( P2 - P1 ) / d
	  x3 = x1 + a * (x2 - x1) / dist;
  	y3 = y1 + a * (y2 - y1) / dist;

	  // two circles intersect usually in 2 points. Find out which one to select
  	if (obj1.dir > obj2.dir)
	  {
		  // result x = x3 + h ( y2 - y1 ) / d
  		m_pos.x = x3 + h * (y2 - y1) / dist;
	  	m_pos.y = y3 - h * (x2 - x1) / dist;
  	}
	  else
  	{
	  	m_pos.x = x3 - h * (y2 - y1) / dist;
		  m_pos.y = y3 + h * (x2 - x1) / dist;
  	}

	  // calculate the absolute direction
	  r1 = obj1.dist;
	  ratio = (y1 - m_pos.y) / r1;
	  int beta = (int) Angle.Rad2Deg(Math.asin(ratio));

	  if (m_pos.x > x1)
		  m_direction = 180 - beta - obj1.dir;
	  else
		  m_direction = beta - obj1.dir;

	  if (m_direction > 180)
		  m_direction -= 360;

		if (m_side == 'r')
		{
			m_pos.x *= -1.0;
			m_pos.y *= -1.0;
      m_direction += ((m_direction > 0) ? -180.0 : 180.0);
		}
	}

	//---------------------------------------------------------------------------
	/** This function calculates current position */
	private void positionFromFlagAndLine(ObjectInfo obj, ObjectInfo line)
	{
		double    tmp;
		Vector2d  obj_pos ;
    double    line_dir, obj_dir;


		obj_pos = (Vector2d)m_landmarks.get(obj.name);
		if(obj_pos == null)
			return;
    line_dir = Angle.Deg2Rad(line.dir);
    obj_dir = obj.dir;

		switch (line.name.charAt(5))
		{
		case 't' :
			m_direction = -line.dir - ((line.dir < 0) ? 180.0 : 0.0) ;
			m_pos.y = -Settings.PITCH_WIDTH/2.0 + Math.abs(line.dist * Math.sin(line_dir)) ;

			if ((tmp = Math.pow(obj.dist,2) - Math.pow(obj_pos.y - m_pos.y,2)) < 0.0)
				tmp *= -1.0 ;
			obj_dir -= m_direction ;
			if (obj_dir > -90.0 && obj_dir < 90.0)
				m_pos.x = obj_pos.x - Math.sqrt(tmp) ;
			else
				m_pos.x = obj_pos.x + Math.sqrt(tmp) ;
			break ;

		case 'b' :
			m_direction = -line.dir - ((line.dir > 0) ? -180.0 : 0.0) ;
			m_pos.y = Settings.PITCH_WIDTH/2.0 - Math.abs(line.dist * Math.sin(line_dir)) ;

			if ((tmp = Math.pow(obj.dist,2) - Math.pow(obj_pos.y - m_pos.y,2)) < 0.0)
				tmp *= -1.0 ;
			obj_dir -= m_direction ;
			if (obj_dir > -90.0 && obj_dir < 90.0)
				m_pos.x = obj_pos.x - Math.sqrt(tmp) ;
			else
				m_pos.x = obj_pos.x + Math.sqrt(tmp) ;
			break ;

		case 'l' :
			m_direction = -line.dir - ((line.dir < 0) ? -90.0 : 90.0) ;
			m_pos.x = -Settings.PITCH_LENGTH/2.0 + Math.abs(line.dist * Math.sin(line_dir)) ;

			if ((tmp = Math.pow(obj.dist,2) - Math.pow(obj_pos.x - m_pos.x,2)) < 0.0)
				tmp *= -1.0 ;
			obj_dir -= m_direction ;
			if (obj_dir < 0.0)
				m_pos.y = obj_pos.y - Math.sqrt(tmp) ;
			else
				m_pos.y = obj_pos.y + Math.sqrt(tmp) ;
			break ;

		case 'r' :
			m_direction = -line.dir - ((line.dir > 0) ? -90.0 : 90.0) ;
			m_pos.x = Settings.PITCH_LENGTH/2.0 - Math.abs(line.dist * Math.sin(line_dir)) ;

			if ((tmp = Math.pow(obj.dist,2) - Math.pow(obj_pos.x - m_pos.x,2)) < 0.0)
				tmp *= -1.0 ;
			obj_dir -= m_direction ;
			if (obj_dir < 0.0)
				m_pos.y = obj_pos.y - Math.sqrt(tmp) ;
			else
				m_pos.y = obj_pos.y + Math.sqrt(tmp) ;
			break ;

		}

		if (m_side == 'r')
		{
			m_direction += ((m_direction > 0) ? -180.0 : 180.0) ;
			m_pos.x *= -1.0 ;
			m_pos.y *= -1.0 ;
		}
  }

  //===========================================================================
  // Private members
  private Body        m_body;
  private String      lastSeeInfo;
	//-----------------------------------
	// Infomation about team
	private char				m_side;
	private String			m_team;
	//-----------------------------------
	// Infomation about field object
  /** Ball info */
	private ObjectInfo	ball = null;
  /** Player info */
	private Vector			m_players = new Vector();
	private Hashtable		m_landmarks = new Hashtable();
	//-----------------------------------
	// Infomation about current position
  private Vector2d  m_pos = new Vector2d();
	private double  	m_direction;						// direction of the player
  /** this flag is used to signal that there is new visual information */
  private long      infoTime = 0;
  /** this flag indicates that we should turn */
  private double    m_turnAngle;
  private int       m_scanAreaID = 0;
}

