/*
  Copyright (C) 2002 by Krzysztof Langner

  This code is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package roboplayer.player;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.FileReader;
import java.net.InetAddress;
import java.net.SocketException;
import java.util.StringTokenizer;
import javax.swing.JFrame;
import roboplayer.body.Settings;
import roboplayer.geometry.Vector2d;
import roboplayer.geometry.Rectangle;
import roboplayer.body.Body;
import roboplayer.memory.Memory;
import roboplayer.query.QuerySystem;
import roboplayer.expert.Expert;
import roboplayer.action.Actions;
import roboplayer.debug.DebugPanel;


public class Player extends Thread
{
	//---------------------------------------------------------------------------
	/** This constructor opens socket for  connection with server */
	public Player(InetAddress host, int port)
		throws SocketException
	{
    body = new Body(host, port);
	}


	//---------------------------------------------------------------------------
	/** initialize body for player */
	public boolean init(String teamName, String playerFile, boolean goalkeeper)
	{
		// first we need to initialize connection with server
		if(!body.init(teamName, goalkeeper))
      return false;

    // Create core modules
    memory = new Memory(body, body.getSide(), teamName);
    // Create high level modules
    Actions actions = new Actions(body, memory);
    QuerySystem queries = new QuerySystem(memory);
    expert = new Expert(actions, queries);
    body.addTickListener(expert);

    readPlayerFile(playerFile);

    return true;
	}


	//---------------------------------------------------------------------------
	/** initialize body for player */
	private void readPlayerFile(String playerFile)
	{
    BufferedReader  in;
    String          line;
    StringTokenizer tokenizer;
    String          key;
    String          value;


    // read info from file
    try
    {
      in = new BufferedReader(new FileReader(playerFile.trim()));
      // In the first line expect team name
      while((line = in.readLine()) != null){
        line.trim();
        // check if this is comment
        if(line.length() > 0 && line.charAt(0) == '#')
          continue;

        tokenizer = new StringTokenizer(line, ":", false);
        // check if this is valid command
        if(tokenizer.countTokens() != 2)
          continue;

        key = tokenizer.nextToken().trim();
        value = tokenizer.nextToken().trim();
        if(key.compareToIgnoreCase("DefaultPos") == 0){
          StringTokenizer t = new StringTokenizer(value, ",", false);
          if(t.countTokens() == 2){
            Vector2d v = new Vector2d(Double.parseDouble(t.nextToken()),
                                      Double.parseDouble(t.nextToken()));
            memory.getTacticalInfo().setDefaultPos(v);
          }
        }
        else if(key.compareToIgnoreCase("DefaultArea") == 0){
          StringTokenizer t = new StringTokenizer(value, ",", false);
          if(t.countTokens() == 4){
            Rectangle area = new Rectangle(
              Double.parseDouble(t.nextToken()),
              Double.parseDouble(t.nextToken()),
              Double.parseDouble(t.nextToken()),
              Double.parseDouble(t.nextToken()));
            memory.getTacticalInfo().setArea(area);
          }
        }
        else if(key.compareToIgnoreCase("Expert") == 0){
          expert.loadFile(value);
        }
        else if(key.compareToIgnoreCase("Debug") == 0){
          JFrame frame = new JFrame();
          frame.setSize(640, 400);
          DebugPanel dp = new DebugPanel(memory);
          body.addTickListener(dp);
          frame.getContentPane().add(dp);
          frame.show();
        }

      }
 		}
    catch(IOException e){
			System.err.println("Error processing file\n");
			System.err.println(e.getMessage());
      System.exit(-1);
    }
  }




	//---------------------------------------------------------------------------
	// This is main loop for player
	public void run()
	{
    // run simulation
    body.run();

    // stop simulation
    body.bye();
    System.out.println("bye");

	}

  //===========================================================================
  // Private members
  /** Core Modules */
  private Memory        memory;
  private Body          body;
  private Expert        expert;

}
